C =====================================================================
C TREECODE: Fortran77 version of hierarchical N-body code (O(N log N)).

C     This program evolves a self-gravitating N-body system, using
C     a hierarchical O(N log N) algorithm to compute gravitational
C     forces (see Barnes, J.E. & Hut, P. 1986, Nature, 324, 446).
C     The computational system of units is determined by the input
C     data, with the assumption that G = 1.  Particles are not
C     required to have identical masses.

C     This code is based on a Fortran implementation of the
C     hierarchical algorithm developed and kindly provided by
C     Lars Hernquist.  Several additional refinments by Jun Makino
C     have been incorporated as well.

C     The present version of the code includes quadrupole moments
C     when calculating the gravitational fields of cells.  It also
C     incorporates an improved opening criterion which increases
C     the accuracy of the forces in certain exceptional situations.

C					Joshua Barnes
C					April 23, 1993
C =====================================================================



C ==============================================================
C TREECD: top-level N-body evolution program.  Its tasks are to:
C  *  read input parameters and initial conditions;
C  *  perform an initial force calculation, reporting
C     system diagnostics and time required;
C  *  advance the system for NSTEPS time-steps, periodically
C     saving a snapshot of the system;
C  *  terminate the simulation and close data files.
C ==============================================================

      PROGRAM TREECD

	INCLUDE 'treedefs.f'
	INTEGER NDIAG, N

C       ---------------------
C       Initialize cpu timer.
C       ---------------------
	CALL SECOND(CPUT0)
C	----------------------------------------------------------------
C	Open data files, read input parameters and initial system state,
C       initialize system parameters, and start output.
C	----------------------------------------------------------------
        CALL INPARM
        CALL GETBDS
        CALL INITCD
        CALL BEGOUT
C       -----------------------------------------------
C	Generate tree and compute initial acceleration.
C	-----------------------------------------------
        CALL MKTREE
	CALL ACCEL
C	------------------------------------------
C	Output timing data and system diagnostics.
C	------------------------------------------
        CALL OUTTIM
        CALL OUTLOG
        CALL DIAGS
C       -----------------------------------------------
C       Set number of steps between diagnostic outputs.
C       -----------------------------------------------
	NDIAG = MAX(1, NSTEPS / 128)
C	=================================================================
C	Primary loop to advance system state for a given number of steps.
C	=================================================================
        DO 100 N = 1, NSTEPS
C	  ------------------------------------------------------
C	  Advance velocity by 1/2 step, positions by 1 timestep.
C	  ------------------------------------------------------
          CALL ADVVEL(DTIME / 2.0)
          CALL ADVPOS(DTIME)
C	  ----------------------------
C	  Create tree, compute forces.
C	  ----------------------------
          CALL MKTREE
          CALL ACCEL
C	  ----------------------------------------------------
C	  Advance velocity 1/2 step to synchronize tvel, tpos.
C	  ----------------------------------------------------
          CALL ADVVEL(DTIME / 2.0)
C	  ----------------------------------------------------
C	  At appropriate intervals, output system diagnostics.
C	  ----------------------------------------------------
          IF (MOD(N, NDIAG) .EQ. 0) THEN
            CALL OUTLOG
            CALL DIAGS
	  ENDIF
C         ----------------------------------------------
C	  At appropriate intervals, output system state.
C         ----------------------------------------------
          IF (MOD(N, NOUT) .EQ. 0)
     &      CALL PUTBDS(.FALSE.)
 100    CONTINUE
C	----------------------------------
C	Output final snapshot with masses.
C	----------------------------------
        CALL PUTBDS(.TRUE.)
C	------------------------------------------------------------------
C	Stop timing, write timing data, close files, terminate simulation.
C	------------------------------------------------------------------
        CALL OUTCPU
        CALL STPOUT
      END

C ----------------------------------------------------------
C INITCD: initialize system parameters that depend on either
C the input data or defined PARAMETERS.
C ----------------------------------------------------------

      SUBROUTINE INITCD
 
	INCLUDE 'treedefs.f'
 
C	---------------------------------------
C	Initialize position and velocity times.
C	---------------------------------------
        TPOS = TNOW
        TVEL = TNOW
C	---------------------------------
C	Initialize box coordinate system.
C	---------------------------------
	RSIZE = 4.0
      END

C ---------------------------------------------------------------
C ADVPOS: advance the positions of the bodies for a timestep DT.
C ---------------------------------------------------------------
 
      SUBROUTINE ADVPOS(DT)
      REAL*8 DT

	INCLUDE 'treedefs.f'
        INTEGER K, P

C	------------------------------------------------- 
C	Loop over all spatial coordinates for all bodies.
C	------------------------------------------------- 
        DO 20 K = 1, NDIM
          DO 10 P = 1, NBODY
            POS(P,K) = POS(P,K) + DT*VEL(P,K)
 10       CONTINUE
 20     CONTINUE
C	----------------------------------
C	Update position time, system time.
C	----------------------------------
        TPOS = TPOS+DT
        TNOW = TPOS
      END

C ----------------------------------------------------------------
C ADVVEL: advance the velocities of the bodies for a timestep dt.
C ----------------------------------------------------------------
 
      SUBROUTINE ADVVEL(DT)
      REAL*8 DT

	INCLUDE 'treedefs.f'
        INTEGER K, P

C	------------------------------------------------- 
C	Loop over all spatial coordinates for all bodies.
C	------------------------------------------------- 
        DO 20 K = 1, NDIM
          DO 10 P = 1, NBODY
            VEL(P,K) = VEL(P,K) + DT*ACC(P,K)
 10       CONTINUE
 20     CONTINUE
C	----------------------------------
C	Update velocity time, system time.
C	----------------------------------
        TVEL = TVEL+DT
        TNOW = TVEL
      END

C ----------------------------------------------------------------
C DIAGS: compute diagnostics for the system: total energy, total
C kinetic energy, total potential energy, angular momentum, center
C of mass coordinates, and center of mass velocity.
C ----------------------------------------------------------------
 
      SUBROUTINE DIAGS

	INCLUDE 'treedefs.f'
        INTEGER P, K
 
C       --------------------------------------------------------------
C       Check that tpos and tvel are synchronized to one part in 1000.
C       --------------------------------------------------------------
        IF (ABS(TPOS - TVEL) .GT. DTIME/1.0E3)
     &     CALL TERROR(' DIAGS: TIMES NOT SYNCHRONIZED')
C       ---------------------------------------------
C       Zero the accumulators for system diagnostics.
C       ---------------------------------------------
        MTOT = 0.0
        EKTOT = 0.0
        EPTOT = 0.0
        DO 10 K = 1, NDIM
          CMPOS(K) = 0.0
          CMVEL(K) = 0.0
          AMVEC(K) = 0.0
 10     CONTINUE
C       ----------------------------------------------------------
C       Loop over bodies to compute net mass and potential energy.
C       ----------------------------------------------------------
        DO 20 P = 1, NBODY
          MTOT = MTOT + MASS(P)
          EPTOT = EPTOT + 0.5 * MASS(P) * PHI(P)
 20     CONTINUE
C       -----------------------------------------------------------------
C       Compute net kinetic energy, center of mass position and velocity.
C       -----------------------------------------------------------------
        DO 30 K = 1, NDIM
          DO 40 P = 1, NBODY
            EKTOT = EKTOT + 0.5 * MASS(P) * VEL(P,K)**2
            CMPOS(K) = CMPOS(K) + MASS(P) * POS(P,K)
            CMVEL(K) = CMVEL(K) + MASS(P) * VEL(P,K)
 40       CONTINUE
          CMVEL(K) = CMVEL(K) / MTOT
          CMPOS(K) = CMPOS(K) / MTOT
 30     CONTINUE
C       ----------------------------
C       Compute total system energy.
C       ----------------------------
        ETOT = EKTOT + EPTOT
C       ---------------------------------------
C       Compute angular momentum of the system.
C       ---------------------------------------
        DO 50 P = 1, NBODY
          AMVEC(1) = AMVEC(1) + MASS(P) *
     &                 (POS(P,2)*VEL(P,3) - POS(P,3)*VEL(P,2))
          AMVEC(2) = AMVEC(2) + MASS(P) *
     &                 (POS(P,3)*VEL(P,1) - POS(P,1)*VEL(P,3))
          AMVEC(3) = AMVEC(3) + MASS(P) *
     &                 (POS(P,1)*VEL(P,2) - POS(P,2)*VEL(P,1))
 50     CONTINUE
C       ----------------------------------
C       Write diagnostics to the log file.
C       ----------------------------------
        CALL OUTDAG
      END

C -------------------------------------------------------------
C TERROR: terminate the program as the result of a fatal error,
C closing the output files and writing timing information.
C -------------------------------------------------------------

      SUBROUTINE TERROR(MSG)
      CHARACTER*(*) MSG

	INCLUDE 'treedefs.f'

C       ------------------------------------
C       Write error message to the log file.
C       ------------------------------------
        CALL OUTERR(MSG)
C       ---------------------------------------------------
C       Output timing data, close files, terminate the run.
C       ---------------------------------------------------
        CALL OUTCPU
        CALL STPOUT
	STOP
      END
