C ==================================
C TREEIO: I/O routines for TREECODE.
C ==================================

C --------------------------------------------------------------------
C INPARM: read in following input parameters:
C
C        hdline    : identification string for the run.
C        nsteps    : number of timesteps.
C        nout      : output system state once every nout steps.
C        dtime     : the timestep.
C        theta     : error tolerance; 0.0 => exact (PP) calculation.
C        eps       : potential softening parameter.
C        usquad    : option to include (.TRUE.) quadrupole terms.

C --------------------------------------------------------------------

      SUBROUTINE INPARM
 
	INCLUDE 'treedefs.f'
        CHARACTER*8 PARSFL
        PARAMETER(PARSFL = 'treepars')
 
C       -----------------------------------------------------
C       Open parameter file, read parameters, close the file.
C       -----------------------------------------------------
        OPEN(UNIT = UPARS, FILE = PARSFL, STATUS = 'OLD')
        READ (UPARS, '(A)') HDLINE

        READ (UPARS, *) NSTEPS, NOUT, DTIME, THETA, EPS, USQUAD

        CLOSE(UNIT = UPARS)
      END

C ---------------------------------------------------------
C GETBDS: read in particle data, assumed to be of the form:
C
C       nbody		: number of bodies.
C       ndim		: number of space dimensions.
C       time		: initial value of time.
C       mass(1)		: mass of first particle.
C       .......	
C       mass(n)		: mass of last particle.
C       x(1) y(1) z(1)	: position of first particle.
C       ..............
C       x(n) y(n) z(n)	: position of last particle.
C       u(1) v(1) w(1)	: velocity of first particle.
C       ..............
C       u(n) v(n) w(n)	: velocity of last particle.
C ---------------------------------------------------------

      SUBROUTINE GETBDS

	INCLUDE 'treedefs.f'
        CHARACTER*8 BODIFL
        PARAMETER(BODIFL = 'treebodi')
        INTEGER INDIM, P, K

C       --------------------
C       Open body data file.
C       --------------------

        OPEN(UNIT = UBODI, FILE = BODIFL, STATUS = 'OLD')

C       ------------------
C       Read in body data.
C       ------------------

        READ (UBODI, *) NBODY
        READ (UBODI, *) INDIM

C       ----------------------------------------------------------
C       Terminate the simulation if NBODY or INDIM not consistent.
C       ----------------------------------------------------------
        IF (NBODY .LE. 0 .OR. NBODY .GT. MXBODY .OR. INDIM .NE. NDIM)
     &    CALL TERROR(' GETBDS: INCONSISTENT OR ILLEGAL INPUTS')
C       --------------------------------------------
C       Read time, masses, positions and velocities.
C       --------------------------------------------

        READ (UBODI, *) TNOW
        DO 10 P = 1, NBODY
          READ (UBODI, *) MASS(P)
  10    CONTINUE
        DO 20 P = 1, NBODY
          READ (UBODI, *) (POS(P,K), K = 1, NDIM)
  20    CONTINUE
        DO 30 P = 1, NBODY
          READ (UBODI, *) (VEL(P,K), K = 1, NDIM)
  30    CONTINUE

C       --------------------
C       Close up input file.
C       --------------------
        CLOSE(UNIT = UBODI)
      END

C ----------------------------------------------------
C BEGOUT: initialize disk files for subsequent output.
C ----------------------------------------------------

      SUBROUTINE BEGOUT 

	INCLUDE 'treedefs.f'
        CHARACTER*8 BODOFL, BODFFL, LGFILE
        PARAMETER(BODOFL = 'treebodo')
        PARAMETER(BODFFL = 'treebodf')
        PARAMETER(LGFILE = 'treelogs')
 
C       --------------------------------
C       Open log file and output header.
C       --------------------------------
        OPEN(UNIT = ULOG, FILE = LGFILE, STATUS = 'NEW')
        CALL OUTHDR
C       ----------------------------
C       Open body data output files.
C       ----------------------------

        OPEN(UNIT = UBODO, FILE = BODOFL, STATUS = 'NEW')
        OPEN(UNIT = UBODF, FILE = BODFFL, STATUS = 'NEW')

      END
 
C ----------------------------------------------------------
C OUTHDR: write header to log file listing input parameters.
C ----------------------------------------------------------

      SUBROUTINE OUTHDR
 
	INCLUDE 'treedefs.f'
 
        WRITE (ULOG, '(/)')
        WRITE (ULOG, '(1X,72(''*''))')
        WRITE (ULOG, '(1X,''*'',70X,''*'')')
        WRITE (ULOG, '(1X,''*'',19X,1A32,19X,''*'')') HDLINE
        WRITE (ULOG, '(1X,''*'',70X,''*'')')
        WRITE (ULOG, '(1X,72(''*''))')
        WRITE (ULOG, '(1X,''*'',70X,''*'')')
        WRITE (ULOG, 30) NBODY, NSTEPS, NOUT, DTIME
 30     FORMAT(1X, '*', 4X, 'NBODY =', 1I7, 4X, 'NSTEPS =', 1I4, 4X,
     &         'NOUT =', 1I4, 4X, 'DTIME =', 1F7.3, 4X, '*')
        WRITE (ULOG, '(1X,''*'',70X,''*'')')

        WRITE (ULOG, 40) USQUAD, EPS, THETA
 40     FORMAT(1X, '*', 4X, 'USEQUAD =', 1L2, 7X, 'EPS =', 1F6.3, 5X,
     &         'THETA =', 1F5.2, 2X, 7X, 3X, 8X, '*')

        WRITE (ULOG, '(1X,''*'',70X,''*'')')

	WRITE (ULOG, '(1X,''*'',4X,1A10,1A30,26X,''*'')')
     &	       'OPTIONS:  ', 'IMPROVED ACCEPTANCE CRITERION '




        WRITE (ULOG, '(1X,''*'',70X,''*'')')
        WRITE (ULOG, '(1X,72(''*''))')
      END

C ---------------------------------------------------
C OUTTIM: write force calculation timing to log file.
C ---------------------------------------------------
 
      SUBROUTINE OUTTIM

	INCLUDE 'treedefs.f'

        WRITE (ULOG, '(/,1X,''     CPU TIME USED:'',3A16)')
     &         'MKTREE', 'ACCEL', 'TOTAL'
        WRITE (ULOG, '(1X,''       (SECONDS)   '',3F16.2)')
     &         CPUMKT, CPUACC, CPUMKT + CPUACC
      END

C --------------------------------------------------------
C OUTLOG: write force calculation diagnostics to log file.
C --------------------------------------------------------

      SUBROUTINE OUTLOG
 
	INCLUDE 'treedefs.f'
        DOUBLE PRECISION CPUT
 
C       -----------------------------
C       Get system-specific cpu time.
C       -----------------------------
        CALL SECOND(CPUT)
        CPUT = CPUT - CPUT0
C       -------------------------------------------------
C       OUTPUT CPU TIME AND FORCE EVALUATION DIAGNOSTICS.
C       -------------------------------------------------
        WRITE (ULOG, '(//)')

        WRITE (ULOG, '(1X,2A9,2A10,2A8,A9)') 'TIME', 'NCELL',
     &        'NBTOT', 'NCTOT', 'NTMAX', 'NTAVG', 'CPUTIME'
        WRITE (ULOG, '(1X,F9.3,I9,2I10,2I8,F9.2)')  TNOW, NCELL,
     &              NBTOT, NCTOT, NTMAX, NTAVG, CPUT/60.0

      END

C -------------------------------------------------
C OUTDAG: write N-body diagnostic data to log file.
C -------------------------------------------------
 
      SUBROUTINE OUTDAG

	INCLUDE 'treedefs.f'
        INTEGER K, IPRINT
        DATA IPRINT /0/

        WRITE (*, '(1X,I4,1X,6(1PE12.4))') IPRINT,
     &              (CMPOS(K), K = 1, NDIM), (CMVEL(K), K = 1, NDIM)
        IPRINT = IPRINT+1
C       ----------------------------------------------------------------
C       Write mass, energy, angular momentum, center of mass quantities.
C       ----------------------------------------------------------------
        WRITE (ULOG, '(/)')
        WRITE (ULOG, '(1X,4A16)') 'ETOT', 'EKIN', 'EPOT', 'MASS'
        WRITE (ULOG, '(1X,3F16.8,F16.8)') ETOT, EKTOT, EPTOT, MTOT
        WRITE (ULOG, '(/)')
        WRITE (ULOG, '(1X,16X,3A16)') 'X', 'Y', 'Z'
        WRITE (ULOG, '(1X, 8X,A8,3E16.6)')
     &         'JTOT', AMVEC(1), AMVEC(2), AMVEC(3)
        WRITE (ULOG, '(1X, 8X,A8,3E16.6)')
     &         'CMPOS', (CMPOS(K), K = 1, NDIM)
        WRITE (ULOG, '(1X, 8X,A8,3E16.6)')
     &         'CMVEL', (CMVEL(K), K = 1, NDIM)
      END

C -----------------------------------------------------------
C OUTERR: output error messages to the log file and terminal.
C -----------------------------------------------------------

      SUBROUTINE OUTERR(MSG)
      CHARACTER*(*) MSG

	INCLUDE 'treedefs.f'
 
C       ------------------------------------------------------
C       Write the message, surrounded by stars for visibility.
C       ------------------------------------------------------
        WRITE (ULOG, '(/,1X,72(''*''))')
        WRITE (ULOG, '(/,A)') MSG
        WRITE (ULOG, '(/,1X,72(''*''))')
C       ---------------------------
C       Repeat message to terminal.
C	---------------------------
        WRITE (UTERM, '(/,1X,72(''*''))')
        WRITE (UTERM, '(/,A)') MSG
        WRITE (UTERM, '(/,1X,72(''*''))')
      END

C -----------------------------------------------
C OUTCPU: output cpu timing data to the log file.
C -----------------------------------------------

      SUBROUTINE OUTCPU

	INCLUDE 'treedefs.f'
        DOUBLE PRECISION CPUT

C       -----------------------------
C       Get system-specific cpu time.
C       -----------------------------
        CALL SECOND(CPUT)
        CPUT = CPUT - CPUT0
        WRITE (ULOG, '(//,'' ELAPSED CPU TIME (SECONDS) : '',F16.2)')
     &              CPUT
      END

C ------------------------------------------------------
C PUTBDS: write body data to output file or final file.
C ------------------------------------------------------







      SUBROUTINE PUTBDS(FINAL)
      LOGICAL FINAL
 
	INCLUDE 'treedefs.f'
 
C       ------------------------------------
C       Decide which file to use for output.
C       ------------------------------------
        IF (FINAL) THEN
C         -------------------------------------------
C         Output final coordinates, including masses.
C         -------------------------------------------

          WRITE (UBODF,  '(I20)') NBODY
          WRITE (UBODF,  '(I20)') NDIM
          WRITE (UBODF,  '(E20.12E2)') TNOW

          CALL PUTMAS(UBODF)
          CALL PUTPOS(UBODF)
          CALL PUTVEL(UBODF)
	ELSE
C	  ---------------------------------
C	  Output particle coordinates only.
C	  ---------------------------------

c          WRITE (UBODO,  '(I20)') NBODY
c          WRITE (UBODO,  '(I20)') NDIM
c          WRITE (UBODO,  '(E20.12E2)') TNOW

          CALL PUTPOS(UBODO)
c          CALL PUTVEL(UBODO)
        ENDIF
      END

C -------------------------------
C PUTMAS: output body mass array.
C -------------------------------

      SUBROUTINE PUTMAS(UBODY)
      INTEGER UBODY

	INCLUDE 'treedefs.f'
        INTEGER P


        DO 10 P = 1, NBODY
          WRITE (UBODY,  '(E20.12E2)') MASS(P)
  10    CONTINUE

      END

C -----------------------------------
C PUTPOS: output body position array.
C -----------------------------------

      SUBROUTINE PUTPOS(UBODY)
      INTEGER UBODY

	INCLUDE 'treedefs.f'
        INTEGER K, P
	
        DO 10 P = 1, NBODY
c          WRITE (UBODY,  '(3E20.12E2)') (POS(P,K), K = 1, NDIM)
           WRITE(UBODY, 999) POS(P,1), POS(P,2), POS(P,3), '  1'
  10    CONTINUE
999     FORMAT(3E12.3E2, A)

      END

C -----------------------------------
C PUTVEL: output body velocity array.
C -----------------------------------

      SUBROUTINE PUTVEL(UBODY)
      INTEGER UBODY

	INCLUDE 'treedefs.f'
        INTEGER K, P


        DO 10 P = 1, NBODY
          WRITE (UBODY,  '(3E20.12E2)') (VEL(P,K), K = 1, NDIM)
  10    CONTINUE

      END
 
C ----------------------------------------------
C STPOUT: terminate output and close open files.
C ----------------------------------------------

      SUBROUTINE STPOUT
 
	INCLUDE 'treedefs.f'
 
        WRITE (ULOG, '(//,1X,''CALCULATION TERMINATED'')')
C       ---------------------
C       Close the open files.
C       ---------------------
        CLOSE(UNIT = UBODO)
        CLOSE(UNIT = UBODF)
        CLOSE(UNIT = ULOG)
      END
